/*  Demonstration for the 3x+1 problem, also called the "Syracuse problem"  */
/*  File:  3x+1.cs							    */

#include <stdio.h>
#include <stdlib.h>
#include "miarith.h"
#include "pmiarith.h"
#include "pmidec.h"
#include "midec.h"

#define NC 16384	/* Number of classes explored in parallel */
    			/* 4 classes is suitable for simulation  */

shape [NC]classes;	/* Classes of numbers == 3 (mod 4)	*/

#define P  4
#define P1 5
#define DP ((30*P + 4)/3)

int:classes	X0[P1],	/* Starting value for current sequence	*/
			X[P1],	/* Current sequence value		*/
            step[P1], 
			One[P1];

int:classes	C,	/* Number of steps in current sequence	*/
			Cmax;	/* Maximun number of steps to stopping	*/

int Ctot;		/* Total steps since starting		*/


/*  Initialize search	*/

void start()
 { with (classes)
    { pint2pmi (X0, P, 4*pcoord(0) + 3);	/* Smallest number in class */
      pmcopy (X, X0,P);  C = Cmax = 0;
      pint2pmi (One, P, (int:current) 1);
/*      printf("One: 0x%08x 0x%08x 0x%08x 0x%08x 0x%08x\n",
                [0]One[0],[0]One[1],[0]One[2],[0]One[3],[0]One[4]); */
      pint2pmi (step,P,(int:current)(4*NC));  Ctot = 0;
      printf("Initialized with %d classes, step = %d\n", NC, 4*NC);
    }
 }


/*  Report all positions at which X == X0.  It is extremely unlikely	*/
/*  that this will ever be used, except for testing.			*/
/*  Call with classes selected, vp's with cycles active (or more).	*/

void report_cycles()
 { int:classes i;	/* Processor coordinates */
   bool:classes cy;    int p;	int Xp[P1];   char conv[80];
   printf("CYCLE(S) DETECTED AT Processor      X0          C\b\n");
   i = pcoord(0);
   cy = (pimcomp (X, X0,P) == 0);	/* True where sequence cycle detected */
   while ( p = (<?= (cy ? i : NC)),	/* Lowest cycle position */
			    p < NC )	/* if any		 */
    { pmp2mp (Xp, X0, P, p);  im2dec (Xp, P, conv, DP);
      printf("Processor:  %7d   Cycles: %10d   X0:\n%s\n",
				     p,  [p]C, conv);
      [p]cy = 0;	/* This item reported */
    }
   printf("\b\n");
  }


/*  Artificial test of report_cycles.  The program is careful to avoid	*/
/*  any net change of the computation state.				*/

void testrc()
 { int:classes Xsv[P1];	/* To save X */
   printf("Test of cycle report -- odd vp's should report cycles.\n");
   with (classes) everywhere
    { pmcopy (Xsv, X, P);  pimadd(X, X0, One, P);
      where ( pcoord(0)&1 ) { pmcopy (X, X0, P);  report_cycles(); }
      pmcopy(X, Xsv, P);			/* So X isn't ultimately changed */
    }
 }

/*  Report mpvar for debugging */

void rep_pmp(char *lbl, int:current V[P1])
 { int p;
   printf("%s:\n", lbl);
   for ( p = 0;  p < NC;  p++ )
   printf("%3d: 0x%08x 0x%08x 0x%08x 0x%08x 0x%08x\n",
            p, [p]V[0], [p]V[1], [p]V[2], [p]V[3], [p]V[4]);
 }


/*  Perform n steps of calculation.				*/
/*  Should be called with classes selected, all vp's active.	*/

void stepper(int n)
 { int k;		/* Loop counter			*/
   int:classes Y[P1];	/* Working parallel variable	*/
   everywhere			/* May help compiler	*/
   for ( k = 0;  k < n;  k++ )
    { pimshift (Y, X,-1,P);	/* X div 2 (integer)  	*/
      where ( X[P]&1 )		/* where X is odd	*/
      { pimadd (X, X, One, P);
        pimadd (X, X, Y,   P);	/*X := 1 + X + Y = (3*x + 1)/2	*/
      }
      else
       pmcopy (X, Y, P);        /*X = Y;    X/2		*/
      C++;			/* Count step		*/
      where ( pimcomp (X, X0, P) == 0)    /*X == X0  Cycle in sequence*/
       if ( |= (int:classes) 1 ) report_cycles();
      where ( pimcomp (X, X0, P) == -1)  /* X < X0 Seqence stops-new sequence*/
       { pimadd (X0, X0, step, P);   	/* X0 = X0 + step  New start   */
	 pmcopy (X, X0, P);     	/***X = X0;  ***/
	 Cmax >?= C;			/* Retain longest sequence	*/
	 C = 0;				/* Reset to count next sequence	*/
       }
    }	/* end for	*/
 }	/* end stepper(n)	*/


/*  Run and time n calculation cycles, report results.	*/

void run(int n)
 { int:classes W[P1];  char conv[80];  int w[P1];
   with (classes) everywhere
    { CM_timer_clear(1);  CM_timer_start(1);
      stepper(n);
      CM_timer_stop(1);
      Ctot += n; 
      pimsub(W,X0,One,P);
      imglbmin(w,W,P);
      im2dec(w,P,conv,DP);
      printf("Confirmed to:\n%s\n longest sequence %d,  %d total steps.",
	           /*  (<?= X0) - 1, */ conv,  (>?= Cmax), Ctot              );
      CM_timer_print(1);
    }
 }


/*  Report state of processor p	*/

void report_proc(int p)
 { int x[P1]; int x0[P1]; char buffx[80], buffx0[80];
   pmp2mp(x,  X,  P, p); im2dec(x,  P, buffx, DP); 
   pmp2mp(x0, X0, P, p); im2dec(x0, P, buffx0,DP);
   printf(
      "Processor:  %7d\nX0:  %s    Cmax: %10d\nX:   %s    C:    %10d\n",
	            p,     buffx0,     [p]Cmax,    buffx,        [p]C    );
/*   printf("X0:  0x%08x, 0x%08x, 0x%08x, 0x%08x, 0x%08x\n",
	  [p]X0[0], [p]X0[1], [p]X0[2], [p]X0[3], [p]X0[4]);
   printf("X:   0x%08x, 0x%08x, 0x%08x, 0x%08x, 0x%08x\n",
	       [p]X[0], [p]X[1], [p]X[2], [p]X[3], [p]X[4]); */
 }


/*  Report complete state of computation.  Works in any situation, but	*/
/*  intended for use with small-scale test calculations.		*/

void report_state()
 { int p;  
   for ( p = 0;  p < NC;  p++ ) report_proc(p);
 }

/*  Main program which manages interactive experiments	*/
/*  Commands are:					*/
/*		  r <n>	- run n steps			*/
/*		  i	- (re)initialize calculation	*/
/*		  s	- report complete state		*/
/*		  S <p> - report state of processor p	*/
/*		  C <p> - check consts in proc. p	*/
/*		  t	- test cycle report		*/
/*		  q	- quit				*/

void main()
 { int ns, p;  char buff[80];
   printf("3x + 1 calculation using %d virtual processors.\n", NC);
   with (classes) everywhere
   while ( printf("3x+1> "),  scanf("%s", buff),  buff[0] != 'q' )
    switch ( buff[0] )
     { case 'r':  scanf("%d", &ns);  
		  if ( [0]X0[P]&1 ) run(ns);      /* [0]X0&1 */
		    else
		     printf("Calculation isn't initialized properly.\n");
		  break;
       case 'i':  start(); break;
       case 's':  report_state(); break;
       case 'S':  scanf("%i", &p);  report_proc(p);  break;
       case 'C':  scanf("%i", &p);
		  printf("One:  0x%08x 0x%08x 0x%08x 0x%08x 0x%08x\n",
		   [p]One[0], [p]One[1], [p]One[2], [p]One[3], [p]One[4]);
		  printf("Step: 0x%08x 0x%08x 0x%08x 0x%08x 0x%08x\n",
		  [p]step[0], [p]step[1], [p]step[2], [p]step[3], [p]step[4]);
		  break;
       case 't':  testrc(); break;
       default:   printf("Bad input -- ignored.\n"); break;
     }
 }
